from __future__ import annotations

import logging
import shutil
from uuid import uuid4
from datetime import timedelta
from pathlib import Path

import fmpy
from fmpy.fmi2 import FMU2Slave
from fmpy.model_description import ModelDescription, ScalarVariable

from cosimtlk.models import FMUInfo, VariableInfo, FMUCausaltyType, FMUInputType

logger = logging.getLogger(__name__)


class FMUSimulator:
    def __init__(
        self,
        path: Path | str,
        step_size: int | timedelta = 1,
        init_values: dict[str, FMUInputType] | None = None,
        output_names: list[str] | None = None,
    ):
        """Simulator using an FMU.

        Args:
            path: Path of the FMU file.
            step_size: FMU simulation step size.
            init_values: Initial values for the inputs.
            output_names: Names of the outputs to be returned. If None, all outputs are returned.
        """
        # Path to the FMU file
        fmu_path = Path(path).resolve()
        if not fmu_path.exists():
            raise FileNotFoundError(f"FMU file not found: {fmu_path}")
        if not fmu_path.suffix == ".fmu":
            raise ValueError(f"FMU file must have .fmu extension: {fmu_path}")
        self.fmu_path = str(fmu_path)

        # Read model description from the FMU
        self.model_description: ModelDescription = fmpy.read_model_description(self.fmu_path)

        # Create input and output maps
        self._inputs: dict[str, ScalarVariable] = {
            variable.name: variable
            for variable in self.model_description.modelVariables
            if variable.causality == FMUCausaltyType.INPUT
        }
        self._outputs: dict[str, ScalarVariable] = {
            variable.name: variable
            for variable in self.model_description.modelVariables
            if variable.causality == FMUCausaltyType.OUTPUT
        }

        # Create FMU slave
        if not self._check_current_platform_is_supported():
            logger.warning(f"{fmpy.platform} is not supported by this FMU, recompiling...")
            fmpy.util.compile_platform_binary(self.fmu_path)
            logger.info(f"Recompiled FMU for platform {fmpy.platform}.")

        self._unzipdir = fmpy.extract(self.fmu_path)
        self.fmu = FMU2Slave(
            unzipDirectory=self._unzipdir,
            guid=self.model_description.guid,
            instanceName=f"{uuid4()}",
            modelIdentifier=self.model_description.coSimulation.modelIdentifier,
        )
        self._instantiated = False
        self.fmu.instantiate(visible=False, callbacks=None, loggingOn=False)
        self._instantiated = True

        # Step size in seconds
        self._current_time = 0
        self._step_size = 0  # reset will set the step size
        self._interested_outputs = output_names
        self.reset(step_size, init_values, output_names)

    def _check_current_platform_is_supported(self) -> bool:
        current_platform = fmpy.platform
        supported_platforms = fmpy.supported_platforms(self.fmu_path)
        return current_platform in supported_platforms

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(t={self._current_time}, path={self.fmu_path!r})"

    def __str__(self) -> str:
        return f"FMU({self.model_description.modelName})"

    def __enter__(self) -> FMUSimulator:
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def close(self) -> None:
        if not self._instantiated:
            return

        self.fmu.terminate()
        self.fmu.freeInstance()
        self._instantiated = False
        shutil.rmtree(self._unzipdir, ignore_errors=True)

    def reset(
        self,
        step_size: int | timedelta = 1,
        init_values: dict[str, FMUInputType] | None = None,
        output_names: list[str] | None = None,
    ) -> None:
        self._check_instantiated()

        # Set step size
        if isinstance(step_size, timedelta):
            step_size = int(step_size.total_seconds())
        self._step_size = step_size

        self._interested_outputs = output_names
        self._initialize_fmu(init_values, current_time=0)

    def _check_instantiated(self):
        if not self._instantiated:
            raise RuntimeError("FMU is not instantiated.")

    def _initialize_fmu(
        self,
        init_values: dict[str, FMUInputType] | None = None,
        current_time: int = 0,
    ):
        self._output_names = {
            "Real": [],
            "Integer": [],
            "Boolean": [],
            "String": [],
        }
        self._output_refs = {
            "Real": [],
            "Integer": [],
            "Boolean": [],
            "String": [],
        }

        interested_outputs = self._interested_outputs or list(self._outputs.keys())
        for variable in self.model_description.modelVariables:
            if (
                variable.causality == FMUCausaltyType.OUTPUT
                and variable.name in interested_outputs
            ):
                self._output_names[variable.type].append(variable.name)
                self._output_refs[variable.type].append(variable.valueReference)

        # Initialize FMU
        self._current_time = current_time
        self.fmu.reset()
        self.fmu.setupExperiment(startTime=self._current_time)

        self.fmu.enterInitializationMode()
        init_values = init_values or {}
        fmpy.simulation.apply_start_values(
            self.fmu, self.model_description, start_values=init_values
        )
        self.fmu.exitInitializationMode()

    def info(self) -> FMUInfo:
        return FMUInfo(
            name=self.model_description.modelName,
            step_size=self._step_size,
            current_time=self._current_time,
        )

    def inputs(self) -> dict[str, VariableInfo]:
        return {
            i.name: VariableInfo(
                type=i.type,
                description=i.description,
                unit=i.unit,
                value=self._read_variable(i),
                min_value=i.min,
                max_value=i.max,
                start_value=i.start,
                variability=i.variability,
            )
            for i in self._inputs.values()
        }

    def outputs(self) -> dict[str, VariableInfo]:
        return {
            o.name: VariableInfo(
                type=o.type,
                description=o.description,
                unit=o.unit,
                value=self._read_variable(o),
                min_value=o.min,
                max_value=o.max,
                start_value=o.start,
                variability=o.variability,
            )
            for o in self._outputs.values()
        }

    def parameters(self) -> dict[str, VariableInfo]:
        return {
            p.name: VariableInfo(
                type=p.type,
                description=p.description,
                unit=p.unit,
                value=self._read_variable(p),
                min_value=p.min,
                max_value=p.max,
                start_value=p.start,
                variability=p.variability,
            )
            for p in self.model_description.modelVariables
            if p.causality == FMUCausaltyType.PARAMETER
        }

    def step(self, n: int = 1, **input_values: FMUInputType) -> dict[str, FMUInputType]:
        self._check_instantiated()
        self._set_inputs(**input_values)

        for _ in range(n):
            self._do_step()

        outputs = self._read_outputs()
        return outputs

    def advance(self, until: int, **input_values: FMUInputType) -> dict[str, FMUInputType]:
        self._check_instantiated()

        if until < self._current_time:
            raise ValueError("Cannot advance time to a time in the past.")

        self._set_inputs(**input_values)

        while self._current_time < until:
            self._do_step()

        outputs = self._read_outputs()
        return outputs

    def _do_step(self):
        self.fmu.doStep(
            currentCommunicationPoint=self._current_time,
            communicationStepSize=self._step_size,
        )
        self._current_time += self._step_size

    def _read_variable(self, variable: ScalarVariable) -> FMUInputType:
        variable_type = variable.type
        variable_reference = variable.valueReference

        if variable_type == "Real":
            return self.fmu.getReal([variable_reference])[0]
        elif variable_type == "Integer":
            return self.fmu.getInteger([variable_reference])[0]
        elif variable_type == "Boolean":
            return self.fmu.getBoolean([variable_reference])[0]
        elif variable_type == "String":
            return self.fmu.getString([variable_reference])[0]
        raise ValueError(
            f"Unknown variable type '{variable_type}' for variable '{variable.name}'."
        )

    def _write_variable(self, variable: ScalarVariable, value: FMUInputType) -> None:
        variable_type = variable.type
        variable_reference = variable.valueReference

        if variable_type == "Real":
            self.fmu.setReal([variable_reference], [float(value)])
        elif variable_type == "Integer":
            self.fmu.setInteger([variable_reference], [int(value)])
        elif variable_type == "Boolean":
            self.fmu.setBoolean([variable_reference], [bool(value)])
        elif variable_type == "String":
            self.fmu.setString([variable_reference], [str(value)])
        else:
            raise ValueError(
                f"Unknown variable type '{variable_type}' for variable '{variable.name}'."
            )

    def _set_inputs(self, **input_values: FMUInputType) -> None:
        for input_name, input_value in input_values.items():
            _input = self._inputs[input_name]
            self._write_variable(_input, input_value)

    def _read_outputs(self) -> dict[str, FMUInputType]:
        outputs = {
            "current_time": self._current_time,
        }
        if self._output_refs["Real"]:
            real_outputs = self.fmu.getReal(self._output_refs["Real"])
            outputs.update(
                {
                    output_name: output_val
                    for output_name, output_val in zip(
                        self._output_names["Real"], real_outputs, strict=True
                    )
                }
            )

        if self._output_refs["Integer"]:
            integer_outputs = self.fmu.getInteger(self._output_refs["Integer"])
            outputs.update(
                {
                    output_name: output_val
                    for output_name, output_val in zip(
                        self._output_names["Integer"], integer_outputs, strict=True
                    )
                }
            )

        if self._output_refs["Boolean"]:
            boolean_outputs = self.fmu.getBoolean(self._output_refs["Boolean"])
            outputs.update(
                {
                    output_name: bool(output_val)
                    for output_name, output_val in zip(
                        self._output_names["Boolean"], boolean_outputs, strict=True
                    )
                }
            )

        if self._output_refs["String"]:
            string_outputs = self.fmu.getString(self._output_refs["String"])
            outputs.update(
                {
                    output_name: str(output_val)
                    for output_name, output_val in zip(
                        self._output_names["String"], string_outputs, strict=True
                    )
                }
            )

        return outputs

    def change_parameters(self, **parameters: FMUInputType) -> None:
        self._initialize_fmu(init_values=parameters, current_time=self._current_time)
        # Set all inputs to their current values
        for input_value in self._inputs.values():
            value = self._read_variable(input_value)
            self._write_variable(input_value, value)



from datetime import datetime
from pathlib import Path
from typing import Any
from uuid import uuid4

from cosimtlk.simulator import FMUSimulator

Record = dict[str, Any]


class SimulatorDB:
    def __init__(self):
        self._db: dict[str, Record] = {}

    def close(self) -> None:
        for key in self._db:
            del self._db[key]

    def create(
        self,
        path: Path,
        **kwargs,
    ) -> Record:
        if not path.exists():
            raise FileNotFoundError(path)

        _id = str(uuid4())
        self._db[_id] = {
            "id": _id,
            "fmu": path.stem,
            "simulator": FMUSimulator(path, **kwargs),
            "created_at": datetime.now(),
        }
        return self.get(_id)

    def list(self) -> list[Record]:
        return [self.get(_id) for _id, simulator_record in self._db.items()]

    def get(self, id: str) -> Record:
        return self._db[id]

    def get_simulator(self, id: str) -> FMUSimulator:
        return self._db[id]["simulator"]

    def delete(self, id: str) -> None:
        del self._db[id]


db = SimulatorDB()



import pytest

from cosimtlk.simulator import FMUSimulator


@pytest.fixture(scope="function")
def simulator():
    path = "/examples/fmus/RealBoolIntTest.fmu"
    simulator = FMUSimulator(path)
    yield simulator
    simulator.close()



from datetime import datetime

from pydantic import BaseModel

from cosimtlk.models import FMUInputType


class SimulatorModel(BaseModel):
    id: str
    fmu: str
    created_at: datetime


class SimulatorInfoModel(BaseModel):
    name: str
    step_size: int
    current_time: int


class VariableInfoModel(BaseModel):
    description: str | None
    unit: str | None
    type: str | None
    value: FMUInputType | None
    start_value: FMUInputType | None
    min_value: FMUInputType | None
    max_value: FMUInputType | None
    variability: str



import logging
from pathlib import Path

from attrs import asdict
from fastapi import APIRouter, Response, Body

from cosimtlk.app.config import settings
from cosimtlk.app.dependencies import db
from cosimtlk.app.schemas import SimulatorModel, SimulatorInfoModel, VariableInfoModel
from cosimtlk.models import FMUInputType

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/simulators", tags=["Simulators"])


@router.get("/", response_model=list[SimulatorModel])
def list_simulators():
    return [SimulatorModel(**simulator) for simulator in db.list()]


@router.post("/", response_model=SimulatorModel)
def create_simulator(
    fmu: str,
    init_values: dict[str, FMUInputType] | None = Body(default=None),
    output_names: list[str] | None = Body(default=None),
    step_size: int = Body(default=1),
):
    fmu_path = Path(settings.fmu_dir).resolve() / f"{fmu}.fmu"
    if not fmu_path.exists():
        return Response(status_code=404)

    simulator = db.create(
        path=fmu_path,
        init_values=init_values,
        output_names=output_names,
        step_size=step_size,
    )
    return SimulatorModel(**simulator)


@router.get("/{id}", response_model=SimulatorModel)
def get_simulator(id: str):
    try:
        return SimulatorModel(**db.get(id))
    except KeyError:
        return Response(status_code=404)


@router.delete("/{id}")
def delete_simulator(
    id: str,
):
    try:
        db.delete(id)
        return Response(status_code=204)
    except KeyError:
        return Response(status_code=404)


@router.get("/{id}/info", response_model=SimulatorInfoModel)
def get_info(id: str):
    simulator = db.get(id)["simulator"]
    return asdict(simulator.info())


@router.get("/{id}/inputs", response_model=dict[str, VariableInfoModel])
def list_inputs(id: str):
    simulator = db.get(id)["simulator"]
    return {k: asdict(v) for k, v in simulator.inputs().items()}


@router.get("/{id}/outputs", response_model=dict[str, VariableInfoModel])
def list_outputs(id: str):
    simulator = db.get(id)["simulator"]
    return {k: asdict(v) for k, v in simulator.outputs().items()}


@router.get("/{id}/parameters", response_model=dict[str, VariableInfoModel])
def list_parameters(id: str):
    simulator = db.get(id)["simulator"]
    return {k: asdict(v) for k, v in simulator.parameters().items()}



import logging
from pathlib import Path

from fastapi import FastAPI

from cosimtlk.app.config import settings
from cosimtlk.app.routers import simulators, simulation

logger = logging.getLogger(__name__)

app = FastAPI(title="FMU Simulator")
app.include_router(simulators.router)
app.include_router(simulation.router)


@app.get("/", tags=["FMUs"])
def list_fmus():
    fmu_dir = Path(settings.fmu_dir).resolve()
    return {
        "path": fmu_dir,
        "fmus": [fmu.name for fmu in fmu_dir.glob("*.fmu")],
    }


@app.on_event("shutdown")
def shutdown_event():
    from cosimtlk.app.dependencies import db

    db.close()



import requests

from cosimtlk.models import SimulatorModel, FMUInputType
from cosimtlk.models import FMUInfo, VariableInfo


class Client:
    def __init__(self, base_url: str):
        self.base_url = base_url
        self.session = requests.Session()

    @classmethod
    def from_parts(cls, host: str = "127.0.0.1", port: int = 8000, secure: bool = False):
        return cls(f"http{'s' if secure else ''}://{host}:{port}")

    def _get(self, path: str, **kwargs) -> dict:
        response = self.session.get(self.base_url + path, **kwargs)
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def _post(self, path: str, body: dict, **kwargs) -> dict:
        response = self.session.post(self.base_url + path, json=body, **kwargs)
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def _put(self, path: str, body: dict, **kwargs) -> dict:
        response = self.session.put(self.base_url + path, json=body, **kwargs)
        if not response.ok:
            response.raise_for_status()
        return response.json()

    def _delete(self, path: str, **kwargs) -> None:
        response = self.session.delete(self.base_url + path, **kwargs)
        if not response.ok:
            response.raise_for_status()
        return None


class SimulatorClient(Client):
    def list_simulators(self) -> list[SimulatorModel]:
        return [SimulatorModel(**simulator) for simulator in self._get("/simulators")]

    def create_simulator(
        self,
        path: str,
        init_values: dict[str, FMUInputType] | None = None,
        output_names: list[str] | None = None,
        step_size: int = 1,
    ) -> SimulatorModel:
        response = self._post(
            "/simulators",
            params=dict(path=path),
            body={
                "init_values": init_values,
                "output_names": output_names,
                "step_size": step_size,
            },
        )
        return SimulatorModel(**response)

    def get_simulator(self, id: str):
        response = self._get(f"/simulators/{id}")
        return SimulatorModel(**response)

    def delete_simulator(self, id: str) -> None:
        return self._delete(f"/simulators/{id}")

    def get_info(self, id: str) -> FMUInfo:
        response = self._get(f"/simulators/{id}/info")
        return FMUInfo(**response)

    def list_inputs(self, id: str) -> dict[str, VariableInfo]:
        response = self._get(f"/simulators/{id}/inputs")
        print(response)
        return {name: VariableInfo(**variable) for name, variable in response.items()}

    def list_outputs(self, id: str) -> dict[str, VariableInfo]:
        response = self._get(f"/simulators/{id}/outputs")
        return {name: VariableInfo(**variable) for name, variable in response.items()}

    def list_parameters(self, id: str) -> dict[str, VariableInfo]:
        response = self._get(f"/simulators/{id}/parameters")
        return {name: VariableInfo(**variable) for name, variable in response.items()}

    def step(
        self, id: str, n: int = 1, input_values: dict[str, FMUInputType] | None = None
    ) -> dict[str, FMUInputType]:
        return self._post(f"/simulators/{id}/step", params=dict(n=n), body=input_values)

    def advance(
        self,
        id: str,
        until: int,
        input_values: dict[str, FMUInputType] | None = None,
    ):
        return self._post(
            f"/simulators/{id}/advance",
            params=dict(until=until),
            body=input_values,
        )

    def change_parameters(
        self,
        id: str,
        parameters: dict[str, FMUInputType],
    ):
        return self._put(
            f"/simulators/{id}/parameters",
            body=parameters,
        )

    def reset(
        self,
        id: str,
        step_size: int = 1,
        output_names: list[str] | None = None,
        init_values: dict[str, FMUInputType] | None = None,
    ):
        return self._put(
            f"/simulators/{id}/reset",
            body={
                "step_size": step_size,
                "output_names": output_names,
                "init_values": init_values,
            },
        )


class RemoteFMUSimulator:
    def __init__(self, client: SimulatorClient, id: str):
        self.client = client
        self.id = id

    def __str__(self):
        return f"RemoteFMUSimulator({self.id})"

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        self.close()

    @classmethod
    def new(
        cls,
        client: SimulatorClient,
        path: str,
        init_values: dict[str, FMUInputType] | None = None,
        output_names: list[str] | None = None,
        step_size: int = 1,
    ):
        simulator = client.create_simulator(path, init_values, output_names, step_size)
        return cls(client, simulator.id)

    def info(self) -> FMUInfo:
        return self.client.get_info(self.id)

    def inputs(self) -> dict[str, VariableInfo]:
        return self.client.list_inputs(self.id)

    def outputs(self) -> dict[str, VariableInfo]:
        return self.client.list_outputs(self.id)

    def parameters(self) -> dict[str, VariableInfo]:
        return self.client.list_parameters(self.id)

    def step(self, n: int = 1, **input_values: FMUInputType) -> dict[str, FMUInputType]:
        return self.client.step(self.id, n, input_values)

    def advance(self, until: int, **input_values: FMUInputType) -> dict[str, FMUInputType]:
        return self.client.advance(self.id, until, input_values)

    def change_parameters(self, **parameters: FMUInputType) -> None:
        self.client.change_parameters(self.id, parameters)

    def reset(
        self,
        step_size: int = 1,
        output_names: list[str] | None = None,
        init_values: dict[str, FMUInputType] | None = None,
    ) -> None:
        self.client.reset(self.id, step_size, output_names, init_values)

    def close(self) -> None:
        self.client.delete_simulator(self.id)



import pytest


def test_init(simulator):
    print(simulator)


def test_closed_raises(simulator):
    simulator.close()
    with pytest.raises(RuntimeError):
        simulator.step()


def test_read_outputs(simulator):
    simulator.reset(init_values={
        'gain.k': 4.0,
        'input_float': 3.14,
        'input_int': 42,
        'input_bool': True,
    })
    outputs = simulator.step(advance_time=False)

    assert outputs == {
        'current_time': 0,
        'outputs': {
            'output_float': 12.56,
            'output_int': 42,
            'output_bool': True,
        }
    }


def test_interested_outputs(simulator):
    simulator.reset(
        output_names=['output_float'],
        init_values={
            'gain.k': 4.0,
            'input_float': 3.14,
            'input_int': 42,
            'input_bool': True,
        }
    )
    outputs = simulator.step(advance_time=False)

    assert outputs == {
        'current_time': 0,
        'outputs': {
            'output_float': 12.56,
        }
    }


def test_step(simulator):
    simulator.reset(init_values={
        'gain.k': 2.0,
    })

    outputs = simulator.step(input_values={
        'input_float': 3.14,
        'input_int': 21,
        'input_bool': False,
    })

    assert outputs == {
        'current_time': 1,
        'outputs': {
            'output_float': 6.28,
            'output_int': 21,
            'output_bool': False,
        }
    }


def test_stepsize(simulator):
    simulator.reset(step_size=5, init_values={
        'gain.k': 2.0,
    })

    outputs = simulator.step(input_values={
        'input_float': 3.14,
        'input_int': 21,
        'input_bool': False,
    })

    assert outputs == {
        'current_time': 5,
        'outputs': {
            'output_float': 6.28,
            'output_int': 21,
            'output_bool': False,
        }
    }


def test_advance(simulator):
    simulator.reset(init_values={
        'gain.k': 2.0,
    })

    outputs = simulator.advance_until(5, input_values={
        'input_float': 3.14,
        'input_int': 21,
        'input_bool': False,
    })

    assert outputs == {
        'current_time': 5,
        'outputs': {
            'output_float': 6.28,
            'output_int': 21,
            'output_bool': False,
        }
    }



import logging

from fastapi import APIRouter, Body
from starlette.responses import JSONResponse

from cosimtlk.app.dependencies import db
from cosimtlk.models import FMUInputType

logger = logging.getLogger(__name__)

router = APIRouter(prefix="/simulators/{id}", tags=["Simulations"])


@router.post("/step")
def step(id: str, n: int = 1, input_values: dict[str, FMUInputType] = Body({})):
    try:
        result = db.get_simulator(id).step(n=n, **input_values)
    except Exception as e:
        logger.exception(e)
        return JSONResponse(status_code=500, content={"error": str(e)})
    return result


@router.post("/advance")
def advance(id: str, until: int, input_values: dict[str, FMUInputType] = Body({})):
    try:
        result = db.get_simulator(id).advance(until, **input_values)
    except ValueError as e:
        logger.exception(e)
        return JSONResponse(status_code=500, content={"error": str(e)})
    return result


@router.put("/parameters")
def change_parameters(id: str, parameters: dict[str, FMUInputType] = Body({})):
    try:
        result = db.get_simulator(id).change_parameters(**parameters)
    except ValueError as e:
        logger.exception(e)
        return JSONResponse(status_code=500, content={"error": str(e)})
    return result


@router.post("/reset")
def reset(
    id: str,
    init_values: dict[str, FMUInputType] = Body(None),
    output_names: list[str] | None = Body(None),
    step_size: int = Body(1),
):
    simulator = db.get(id)["simulator"]
    try:
        simulator.reset(step_size=step_size, output_names=output_names, init_values=init_values)
    except Exception as e:
        logger.exception(e)
        return JSONResponse(status_code=500, content={"error": str(e)})
    return JSONResponse(status_code=200, content={"message": "Success"})



from datetime import datetime
from enum import Enum

from attr import define

FMUInputType = float | int | str | bool


@define
class SimulatorModel:
    id: str
    path: str
    created_at: datetime


@define
class FMUInfo:
    name: str
    step_size: int
    current_time: int


@define
class VariableInfo:
    description: str | None
    value: FMUInputType
    unit: str | None
    type: str | None
    start_value: FMUInputType | None
    min_value: FMUInputType | None
    max_value: FMUInputType | None
    variability: str


class FMUCausaltyType(str, Enum):
    INPUT = "input"
    OUTPUT = "output"
    PARAMETER = "parameter"
    CALCULATED_PARAMETER = "calculatedParameter"
    LOCAL = "local"
